//============================================================================
// camera.hpp : OPENGL-style camera class definition
//  Defines and stores a "camera" composed of 3 coord axes, an origin for
//  these axes (center-of-projection or eye location), a near and far plane
//  as distances from the origin projected along the viewing direction, 
//  the viewplane window extents (projection window defined in cam coords 
//  on the viewplane - near plane). ViewPlane is defined as the near plane.
//  Viewing direction is always defined along the -Z axis with eye at origin.
//============================================================================

#ifndef CAMERA
#define CAMERA

#include <stdio.h>
#include <vec3f.hpp>
#include <mat16fv.hpp>

class Camera
{
 public:

  Vec3f X, Y, Z;            // NORMALIZED CAMERA COORDINATE-AXIS VECTORS
  Vec3f Orig;               // LOCATION OF ORIGIN OF CAMERA SYSTEM IN WORLD COORDS
  float wL,wR,wB,wT;        // WINDOW EXTENTS DEFINED AS A RECT ON NearPlane
  float Near,Far;           // DISTANCES TO NEAR AND FAR PLANES (IN VIEWING DIR)

  Camera();
  Camera(Camera &Cam);
  void Copy(Camera &Cam);

  void LookAt(Vec3f Eye, Vec3f ViewRefPt, Vec3f ViewUp);
  void Perspective(float Yfov, float Aspect, float Ndist, float Fdist);
  void Frustum(float l, float r, float b, float t, float Ndist, float Fdist);

  void TightlyFitToSphere(Vec3f Eye, Vec3f ViewUp, Vec3f Cntr, float Rad);

  void GetLookAtParams(Vec3f *Eye, Vec3f *ViewRefPt, Vec3f *ViewUp);
  void GetPerspectiveParams(float *Yfov, float *Aspect, float *Ndist, float *Fdist);
  void GetFrustumParams(float *l, float *r, float *b, float *t, float *Ndist, float *Fdist);
  Vec3f wCOP();           // WORLD COORDINATE CENTER-OF-PROJECTION (EYE)
  Vec3f ViewDir();        // VIEWING DIRECTION
  Vec3f ViewDirOffAxis();

  float* GetXform_Screen2Obj(float* M, int WW, int WH);
  float* GetXform_Obj2Screen(float* M, int WW, int WH);

  float* GetModelviewMatrix(float* M);
  float* GetProjectionMatrix(float* M);
  float* GetViewportMatrix(float* M, int WW, int WH);

  float* GetInvModelviewMatrix(float* M);
  float* GetInvProjectionMatrix(float* M);
  float* GetInvViewportMatrix(float* M, int WW, int WH);

  Vec3f WorldToCam(Vec3f wP);
  Vec3f CamToWorld(Vec3f cP);

  void LoadIdentityXform();
  void Xform(float M[16]);

  void GetPixelRay(float sx, float sy, int ww, int wh, Vec3f *Start, Vec3f *Dir);

  void WriteToFile(FILE *fp);
  int ReadFromFile(FILE *fp);  // RETURNS "1" IF SUCCESSFUL, "0" IF EOF

  void CalcVerts(Vec3f *V);    // CALCS EIGHT CORNERS OF VIEW-FRUSTUM
  void Print();
  void Display();
};

#endif
