//============================================================================
// mat44.hpp : 4x4 OpenGL-style matrix class
//============================================================================

#include "mat44.hpp"

//---------------------------------------------------------------------------
// CONSTRUCTORS
//---------------------------------------------------------------------------
Mat44::Mat44()
{ 
  Identity(); 
}

Mat44::Mat44(float *N)
{
  M[0]=N[0]; M[4]=N[4];  M[8]=N[8];  M[12]=N[12];
  M[1]=N[1]; M[5]=N[5];  M[9]=N[9];  M[13]=N[13];
  M[2]=N[2]; M[6]=N[6]; M[10]=N[10]; M[14]=N[14];
  M[3]=N[3]; M[7]=N[7]; M[11]=N[11]; M[15]=N[15];
}

Mat44::Mat44(float M0, float M4, float M8, float M12,
      float M1, float M5, float M9, float M13,
      float M2, float M6, float M10, float M14,
      float M3, float M7, float M11, float M15)
{
  M[0]=M0; M[4]=M4;  M[8]=M8;  M[12]=M12;
  M[1]=M1; M[5]=M5;  M[9]=M9;  M[13]=M13;
  M[2]=M2; M[6]=M6; M[10]=M10; M[14]=M14;
  M[3]=M3; M[7]=M7; M[11]=M11; M[15]=M15;
}

//---------------------------------------------------------------------------
// MATRIX/MATRIX AND MATRIX/VECTOR OPERATORS
//---------------------------------------------------------------------------
Mat44& Mat44::operator = (const Mat44& A)      // ASSIGNMENT (=)
{
  M[0]=A.M[0]; M[4]=A.M[4];  M[8]=A.M[8];  M[12]=A.M[12];
  M[1]=A.M[1]; M[5]=A.M[5];  M[9]=A.M[9];  M[13]=A.M[13];
  M[2]=A.M[2]; M[6]=A.M[6]; M[10]=A.M[10]; M[14]=A.M[14];
  M[3]=A.M[3]; M[7]=A.M[7]; M[11]=A.M[11]; M[15]=A.M[15];
  return(*this);
}

Mat44& Mat44::operator = (const float* a) {
  for (int i=0;i<16;i++) {
    M[i] = a[i];
  }
  return *this;
}

Mat44 Mat44::operator * (const Mat44& A) const  // MULTIPLICATION (*)
{
  Mat44 NewM( M[0]*A.M[0] + M[4]*A.M[1] + M[8]*A.M[2] + M[12]*A.M[3],      // ROW 1
              M[0]*A.M[4] + M[4]*A.M[5] + M[8]*A.M[6] + M[12]*A.M[7],
              M[0]*A.M[8] + M[4]*A.M[9] + M[8]*A.M[10] + M[12]*A.M[11],
              M[0]*A.M[12] + M[4]*A.M[13] + M[8]*A.M[14] + M[12]*A.M[15],

              M[1]*A.M[0] + M[5]*A.M[1] + M[9]*A.M[2] + M[13]*A.M[3],      // ROW 2
              M[1]*A.M[4] + M[5]*A.M[5] + M[9]*A.M[6] + M[13]*A.M[7],
              M[1]*A.M[8] + M[5]*A.M[9] + M[9]*A.M[10] + M[13]*A.M[11],
              M[1]*A.M[12] + M[5]*A.M[13] + M[9]*A.M[14] + M[13]*A.M[15],

              M[2]*A.M[0] + M[6]*A.M[1] + M[10]*A.M[2] + M[14]*A.M[3],     // ROW 3
              M[2]*A.M[4] + M[6]*A.M[5] + M[10]*A.M[6] + M[14]*A.M[7],
              M[2]*A.M[8] + M[6]*A.M[9] + M[10]*A.M[10] + M[14]*A.M[11],
              M[2]*A.M[12] + M[6]*A.M[13] + M[10]*A.M[14] + M[14]*A.M[15],

              M[3]*A.M[0] + M[7]*A.M[1] + M[11]*A.M[2] + M[15]*A.M[3],     // ROW 4
              M[3]*A.M[4] + M[7]*A.M[5] + M[11]*A.M[6] + M[15]*A.M[7],
              M[3]*A.M[8] + M[7]*A.M[9] + M[11]*A.M[10] + M[15]*A.M[11],
              M[3]*A.M[12] + M[7]*A.M[13] + M[11]*A.M[14] + M[15]*A.M[15] );
  return(NewM);
}

Vec3f Mat44::operator * (const Vec3f& V) const  // MAT-VECTOR MULTIPLICATION (*) W/ PERSP DIV
{
  float W = M[3]*V.x + M[7]*V.y + M[11]*V.z + M[15];
  Vec3f NewV( (M[0]*V.x + M[4]*V.y + M[8]*V.z  + M[12]) / W,
              (M[1]*V.x + M[5]*V.y + M[9]*V.z  + M[13]) / W,
              (M[2]*V.x + M[6]*V.y + M[10]*V.z + M[14]) / W );
  return(NewV);
}

Vec4f Mat44::operator * (const Vec4f& V) const  // MAT-VECTOR MULTIPLICATION (*)
{
  Vec4f NewV;
  NewV.x = M[0]*V.x + M[4]*V.y + M[8]*V.z  + M[12]*V.w;
  NewV.y = M[1]*V.x + M[5]*V.y + M[9]*V.z  + M[13]*V.w;
  NewV.z = M[2]*V.x + M[6]*V.y + M[10]*V.z + M[14]*V.w;
  NewV.w = M[3]*V.x + M[7]*V.y + M[11]*V.z + M[15]*V.w;
  return(NewV);
}

Mat44::operator const float*() const
{
  return M;
}

Mat44::operator float*()
{
  return M;
}

float& Mat44::operator()(int col, int row)
{
  return M[4*col+row];
}

const float& Mat44::operator()(int col, int row) const
{
  return M[4*col+row];
}

void Mat44::Set(const float* a)
{
  for (int i=0;i<16;i++) {
    M[i] = a[i];
  }
  
}

//---------------------------------------------------------------------------
// Standard Matrix Operations
//---------------------------------------------------------------------------
void Mat44::Identity()
{
  M[0]=M[5]=M[10]=M[15]=1;
  M[1]=M[2]=M[3]=M[4]=M[6]=M[7]=M[8]=M[9]=M[11]=M[12]=M[13]=M[14]=0;
}

void Mat44::Transpose()
{
  float t;
  SWAP(M[1],M[4],t);
  SWAP(M[2],M[8],t);
  SWAP(M[6],M[9],t);
  SWAP(M[3],M[12],t);
  SWAP(M[7],M[13],t);
  SWAP(M[11],M[14],t);
}

//---------------------------------------------------------------------------
// Standard Matrix Affine Transformations
//---------------------------------------------------------------------------
void Mat44::Translate(float Tx, float Ty, float Tz)
{
  M[0]=1; M[4]=0;  M[8]=0;  M[12]=Tx;
  M[1]=0; M[5]=1;  M[9]=0;  M[13]=Ty;
  M[2]=0; M[6]=0;  M[10]=1; M[14]=Tz;
  M[3]=0; M[7]=0;  M[11]=0; M[15]=1;
}

void Mat44::Translate(const Vec3f& T)
{
  M[0]=1; M[4]=0;  M[8]=0;  M[12]=T.x;
  M[1]=0; M[5]=1;  M[9]=0;  M[13]=T.y;
  M[2]=0; M[6]=0;  M[10]=1; M[14]=T.z;
  M[3]=0; M[7]=0;  M[11]=0; M[15]=1;
}

void Mat44::invTranslate(float Tx, float Ty, float Tz)
{
  M[0]=1; M[4]=0;  M[8]=0;  M[12]=-Tx;
  M[1]=0; M[5]=1;  M[9]=0;  M[13]=-Ty;
  M[2]=0; M[6]=0;  M[10]=1; M[14]=-Tz;
  M[3]=0; M[7]=0;  M[11]=0; M[15]=1;
}

void Mat44::invTranslate(const Vec3f& T)
{
  M[0]=1; M[4]=0;  M[8]=0;  M[12]=-T.x;
  M[1]=0; M[5]=1;  M[9]=0;  M[13]=-T.y;
  M[2]=0; M[6]=0;  M[10]=1; M[14]=-T.z;
  M[3]=0; M[7]=0;  M[11]=0; M[15]=1;
}

void Mat44::Scale(float Sx, float Sy, float Sz)
{
  M[0]=Sx; M[4]=0;  M[8]=0;   M[12]=0;
  M[1]=0;  M[5]=Sy; M[9]=0;   M[13]=0;
  M[2]=0;  M[6]=0;  M[10]=Sz; M[14]=0;
  M[3]=0;  M[7]=0;  M[11]=0;  M[15]=1;
}

void Mat44::Scale(const Vec3f& S)
{
  M[0]=S.x; M[4]=0;   M[8]=0;    M[12]=0;
  M[1]=0;   M[5]=S.y; M[9]=0;    M[13]=0;
  M[2]=0;   M[6]=0;   M[10]=S.z; M[14]=0;
  M[3]=0;   M[7]=0;   M[11]=0;   M[15]=1;
}

void Mat44::invScale(float Sx, float Sy, float Sz)
{
  M[0]=1/Sx; M[4]=0;    M[8]=0;     M[12]=0;
  M[1]=0;    M[5]=1/Sy; M[9]=0;     M[13]=0;
  M[2]=0;    M[6]=0;    M[10]=1/Sz; M[14]=0;
  M[3]=0;    M[7]=0;    M[11]=0;    M[15]=1;
}

void Mat44::invScale(const Vec3f& S)
{
  M[0]=1/S.x; M[4]=0;     M[8]=0;      M[12]=0;
  M[1]=0;     M[5]=1/S.y; M[9]=0;      M[13]=0;
  M[2]=0;     M[6]=0;     M[10]=1/S.z; M[14]=0;
  M[3]=0;     M[7]=0;     M[11]=0;     M[15]=1;
}

void Mat44::Rotate(float DegAng, Vec3f Axis)
{
  float RadAng = DegAng*TORADS;
  float ca=(float)cos(RadAng),
        sa=(float)sin(RadAng);
  if (Axis.x==1 && Axis.y==0 && Axis.z==0)  // ABOUT X-AXIS
  {
   M[0]=1; M[4]=0;  M[8]=0;   M[12]=0;
   M[1]=0; M[5]=ca; M[9]=-sa; M[13]=0;
   M[2]=0; M[6]=sa; M[10]=ca; M[14]=0;
   M[3]=0; M[7]=0;  M[11]=0;  M[15]=1;
  }
  if (Axis.x==0 && Axis.y==1 && Axis.z==0)  // ABOUT Y-AXIS
  {
   M[0]=ca;  M[4]=0; M[8]=sa;  M[12]=0;
   M[1]=0;   M[5]=1; M[9]=0;   M[13]=0;
   M[2]=-sa; M[6]=0; M[10]=ca; M[14]=0;
   M[3]=0;   M[7]=0; M[11]=0;  M[15]=1;
  }
  if (Axis.x==0 && Axis.y==0 && Axis.z==1)  // ABOUT Z-AXIS
  {
   M[0]=ca; M[4]=-sa; M[8]=0;  M[12]=0;
   M[1]=sa; M[5]=ca;  M[9]=0;  M[13]=0;
   M[2]=0;  M[6]=0;   M[10]=1; M[14]=0;
   M[3]=0;  M[7]=0;   M[11]=0; M[15]=1;
  }
  else                                      // ARBITRARY AXIS
  {
   Axis.Normalize();
   float x=Axis.x, y=Axis.y, z=Axis.z;
   float x2=x*x, y2=y*y, z2=z*z;
   M[0]=x2+ca*(1-x2); M[4]=(x*y)+ca*(-x*y)+sa*(-z); M[8]=(x*z)+ca*(-x*z)+sa*y;
   M[1]=(x*y)+ca*(-x*y)+sa*z; M[5]=y2+ca*(1-y2); M[9]=(y*z)+ca*(-y*z)+sa*(-x);
   M[2]=(x*z)+ca*(-x*z)+sa*(-y); M[6]=(y*z)+ca*(-y*z)+sa*x; M[10]=z2+ca*(1-z2);
   M[12]=M[13]=M[14]=M[3]=M[7]=M[11]=0;
   M[15]=1;
  }
}

void Mat44::invRotate(float DegAng, Vec3f Axis)
{
  Rotate(DegAng,Axis);
  Transpose();
}

//---------------------------------------------------------------------------
// Same as glFrustum() : Perspective transformation matrix defined by a
//  truncated pyramid viewing frustum that starts at the origin (eye)
//  going in the -Z axis direction (viewer looks down -Z)
//  with the four pyramid sides passing through the sides of a window
//  defined through x=l, x=r, y=b, y=t on the viewplane at z=-n.
//  The top and bottom of the pyramid are truncated by the near and far
//  planes at z=-n and z=-f. A 4-vector (x,y,z,w) inside this frustum
//  transformed by this matrix will have x,y,z values in the range
//  [-w,w]. Homogeneous clipping is applied to restrict (x,y,z) to
//  this range. Later, a perspective divide by w will result in an NDC
//  coordinate 3-vector of the form (x/w,y/w,z/w,w/w)=(x',y',z',1) where
//  x', y', and z' all are in the range [-1,1]. Perspectively divided z'
//  will be in [-1,1] with -1 being the near plane and +1 being the far.
//---------------------------------------------------------------------------
void Mat44::Frustum(float l, float r, float b, float t, float n, float f)
{
  M[0]=(2*n)/(r-l); M[4]=0;           M[8]=(r+l)/(r-l);   M[12]=0;
  M[1]=0;           M[5]=(2*n)/(t-b); M[9]=(t+b)/(t-b);   M[13]=0;
  M[2]=0;           M[6]=0;           M[10]=-(f+n)/(f-n); M[14]=(-2*f*n)/(f-n);
  M[3]=0;           M[7]=0;           M[11]=-1;           M[15]=0;
}

void Mat44::invFrustum(float l, float r, float b, float t, float n, float f)
{
  M[0]=(r-l)/(2*n); M[4]=0;           M[8]=0;               M[12]=(r+l)/(2*n);
  M[1]=0;           M[5]=(t-b)/(2*n); M[9]=0;               M[13]=(t+b)/(2*n);
  M[2]=0;           M[6]=0;           M[10]=0;              M[14]=-1;
  M[3]=0;           M[7]=0;           M[11]=-(f-n)/(2*f*n); M[15]=(f+n)/(2*f*n);
}

//---------------------------------------------------------------------------
// Same as gluPerspective : calls Frustum()
//---------------------------------------------------------------------------
void Mat44::Perspective(float Yfov, float Aspect, float Ndist, float Fdist)
{
  Yfov *= 0.0174532f;  // CONVERT TO RADIANS
  float wT=tanf(Yfov*0.5f)*Ndist, wB=-wT;
  float wR=wT*Aspect, wL=-wR;
  Frustum(wL,wR,wB,wT,Ndist,Fdist);
}

void Mat44::invPerspective(float Yfov, float Aspect, float Ndist, float Fdist)
{
  Yfov *= 0.0174532f;  // CONVERT TO RADIANS
  float wT=tanf(Yfov*0.5f)*Ndist, wB=-wT;
  float wR=wT*Aspect, wL=-wR;
  invFrustum(wL,wR,wB,wT,Ndist,Fdist);
}

//---------------------------------------------------------------------------
// OpenGL VIEWPORT XFORM MATRIX : given Window width and height in pixels
// Transforms the x,y,z NDC values in [-1,1] to (x',y') pixel values and
// normalized z' in [0,1] (near and far respectively).
//---------------------------------------------------------------------------
void Mat44::Viewport(int WW, int WH)
{
  float WW2=(float)WW*0.5f, WH2=(float)WH*0.5f;
  M[0]=WW2;  M[4]=0;     M[8]=0;    M[12]=WW2;
  M[1]=0;    M[5]=WH2;   M[9]=0;    M[13]=WH2;
  M[2]=0;    M[6]=0;     M[10]=0.5; M[14]=0.5;
  M[3]=0;    M[7]=0;     M[11]=0;   M[15]=1;
}

void Mat44::invViewport(int WW, int WH)
{
  float WW2=2.0f/(float)WW, WH2=2.0f/(float)WH;
  M[0]=WW2;  M[4]=0;     M[8]=0;    M[12]=-1.0;
  M[1]=0;    M[5]=WH2;   M[9]=0;    M[13]=-1.0;
  M[2]=0;    M[6]=0;     M[10]=2.0; M[14]=-1.0;
  M[3]=0;    M[7]=0;     M[11]=0;   M[15]=1;
}

//---------------------------------------------------------------------------
// Same as gluLookAt()
//---------------------------------------------------------------------------
void Mat44::LookAt(Vec3f Eye, Vec3f LookAtPt, Vec3f ViewUp)
{
  Vec3f Z = Eye-LookAtPt;  Z.Normalize(); // CALC CAM AXES ("/" IS CROSS-PROD)
  Vec3f X = ViewUp/Z;      X.Normalize();
  Vec3f Y = Z/X;           Y.Normalize();
  Vec3f Tr = -Eye;
  M[0]=X.x;  M[4]=X.y;   M[8]=X.z; M[12]=X*Tr;  // TRANS->ROT
  M[1]=Y.x;  M[5]=Y.y;   M[9]=Y.z; M[13]=Y*Tr;
  M[2]=Z.x;  M[6]=Z.y;  M[10]=Z.z; M[14]=Z*Tr;
  M[3]=0;    M[7]=0;    M[11]=0;   M[15]=1;
}

void Mat44::invLookAt(Vec3f Eye, Vec3f LookAtPt, Vec3f ViewUp)
{
  Vec3f Z = Eye-LookAtPt;  Z.Normalize(); // CALC CAM AXES ("/" IS CROSS-PROD)
  Vec3f X = ViewUp/Z;      X.Normalize();
  Vec3f Y = Z/X;           Y.Normalize();
  M[0]=X.x;  M[4]=Y.x;   M[8]=Z.x; M[12]=Eye.x;  // ROT->TRANS
  M[1]=X.y;  M[5]=Y.y;   M[9]=Z.y; M[13]=Eye.y;
  M[2]=X.z;  M[6]=Y.z;  M[10]=Z.z; M[14]=Eye.z;
  M[3]=0;    M[7]=0;    M[11]=0;   M[15]=1;
}

//---------------------------------------------------------------------------
// VIEWPORT TRANFORMATION MATRIX : given Window width and height in pixels
// FROM "JIM BLINN'S CORNER" JULY '91. This function transforms an
// 4-vector in homogeneous space of the form (x,y,z,w) where
// -w<=x,y,z<=w into (x',y',z',1) where (x',y') is the vectors
// projected pixel location in ([0,WW-1],[0,WH-1]) and z' is the
// normalized depth value mapped into [0,1]; 0 is the near plane
// and 1 is the far plane. VIEWPORT_TOL is introduced to have correct
// mappings. (-1,-1,?,?) in NDC refers to the bottom-left of the
// viewplane window and (0,0,?,?) in screen space refers to the
// bottom-left pixel.
//---------------------------------------------------------------------------
void Mat44::Viewport2(int WW, int WH)
{
  float WW2=(WW-VIEWPORT_TOL)*0.5f, WH2=(WH-VIEWPORT_TOL)*0.5f;
  M[0]=WW2;  M[4]=0;     M[8]=0;    M[12]=WW2;
  M[1]=0;    M[5]=WH2;   M[9]=0;    M[13]=WH2;
  M[2]=0;    M[6]=0;     M[10]=0.5; M[14]=0.5;
  M[3]=0;    M[7]=0;     M[11]=0;   M[15]=1;
}

void Mat44::invViewport2(int WW, int WH)
{
  float WW2=2.0f/(WW-VIEWPORT_TOL), WH2=2.0f/(WH-VIEWPORT_TOL);
  M[0]=WW2;  M[4]=0;     M[8]=0;    M[12]=-1.0;
  M[1]=0;    M[5]=WH2;   M[9]=0;    M[13]=-1.0;
  M[2]=0;    M[6]=0;     M[10]=2.0; M[14]=-1.0;
  M[3]=0;    M[7]=0;     M[11]=0;   M[15]=1;
}

//---------------------------------------------------------------------------
// Handy matrix printing routine.
//---------------------------------------------------------------------------
void Mat44::Print()
{
  printf("\n%f %f %f %f\n",M[0],M[4],M[8],M[12]);
  printf("%f %f %f %f\n",M[1],M[5],M[9],M[13]);
  printf("%f %f %f %f\n",M[2],M[6],M[10],M[14]);
  printf("%f %f %f %f\n\n",M[3],M[7],M[11],M[15]);
}

//---------------------------------------------------------------------------
// Copy contents of matrix into matrix array.
//---------------------------------------------------------------------------
void Mat44::CopyInto(float *Mat)
{
  Mat[0]=M[0]; Mat[4]=M[4];  Mat[8]=M[8];  Mat[12]=M[12];
  Mat[1]=M[1]; Mat[5]=M[5];  Mat[9]=M[9];  Mat[13]=M[13];
  Mat[2]=M[2]; Mat[6]=M[6]; Mat[10]=M[10]; Mat[14]=M[14];
  Mat[3]=M[3]; Mat[7]=M[7]; Mat[11]=M[11]; Mat[15]=M[15];
}

