//============================================================================
// mat44.hpp : 4x4 OpenGL-style matrix class
//============================================================================

#ifndef _MAT44_
#define _MAT44_

#include "vec3f.hpp"
#include "vec4f.hpp"

#define TORADS 0.0174532f
#define SWAP(a,b,t) (t)=(a);(a)=(b);(b)=(t);
#define VIEWPORT_TOL 0.001f

//----------------------------------------------------------------------------
// M[16] = [ 0  4  8 12 ]   |  16 floats was used instead of the normal [4][4]
//         [ 1  5  9 13 ]   |  to be compliant with OpenGL. OpenGL uses
//         [ 2  6 10 14 ]   |  premultiplication with column vectors. These
//         [ 3  7 11 15 ]   |  matrices can be fed directly into the OpenGL
//                          |  matrix stack with glLoadMatrix or glMultMatrix.
//
// [ x' y' z' w' ] = [ 0  4  8 12 ]   [ x ]
//                   [ 1  5  9 13 ] * [ y ] 
//                   [ 2  6 10 14 ]   [ z ]
//                   [ 3  7 11 15 ]   [ w ]
// 
// Loading a [4][4] format matrix directly into the matrix stack (assuming
// premult/col vecs) results in a transpose matrix. M[0]=M[0][0], but
// M[1]!=M[1][0] since M[0][1] would cast to M[1].
//
// However, if we assumed postmult/row vectors we could use [4][4] format,
// but all transformations in this module would be transposed.
//----------------------------------------------------------------------------
class Mat44
{
  public:
    float M[16];  // 0,1,2,3 = 1st col; 4,5,6,7 = 2nd col; etc.

    Mat44();
    Mat44(float *N);
    Mat44(float M0, float M4, float M8, float M12,
          float M1, float M5, float M9, float M13,
          float M2, float M6, float M10, float M14,
          float M3, float M7, float M11, float M15);
    Mat44& operator = (const Mat44& A);       // ASSIGNMENT (=)
    Mat44& operator = (const float* a);       // ASSIGNMENT (=) FROM AN ARRAY OF FLOATS 
    Mat44 operator * (const Mat44& A) const;  // MULTIPLICATION (*)
    Vec3f operator * (const Vec3f& V) const;  // MAT-VECTOR MULTIPLICATION (*) W/ PERSP DIV
    Vec4f operator * (const Vec4f& V) const;  // MAT-VECTOR MULTIPLICATION (*)
////////////// NOT IMPLEMENTED YET. ANY TAKERS?
//    friend Vec3f operator * (const Vec3f& V, const Mat44& M); // MAT-VECTOR PRE-MULTIPLICATON (*) W/ PERP DIV
//    friend Vec4f operator * (const Vec4f& V, const Mat44& M); // MAT-VECTOR PRE-MULTIPLICATON (*)
//    Mat44 operator * (float a) const;         // SCALAR POST-MULTIPLICATION
//    friend Mat44 operator * (float a, const Mat44& M) const; // SCALAR PRE-MULTIPLICATION
//    Mat44 operator / (float a) const;         // SCALAR DIVISION
//    Mat44 operator + (Mat44& M) const;        // ADDITION (+)
//    Mat44& operator += (Mat44& M);            // ACCUMULATE ADD (+=)
//    Mat44& operator *= (float a);             // ACCUMULATE MULTIPLY (*=)
//    Mat44& operator /= (float a);             // ACCUMULATE DIVIDE (/=)
//    bool Invserse();                          // MATRIX INVERSE
////////////// NOT IMPLEMENTED YET. ANY TAKERS?

    operator const float*() const;            // CAST TO A FLOAT ARRAY
    operator float*();                        // CAST TO A FLOAT ARRAY
    float& operator()(int col, int row);      // 2D ARRAY ACCESSOR
    const float& operator()(int col, int row) const;  // 2D ARRAY ACCESSOR
    void Set(const float* a);                 // SAME AS ASSIGNMENT (=) FROM AN ARRAY OF FLOATS

    void Identity();
    void Transpose();
    void Translate(float Tx, float Ty, float Tz);
    void Translate(const Vec3f& T);
    void invTranslate(float Tx, float Ty, float Tz);
    void invTranslate(const Vec3f& T);
    void Scale(float Sx, float Sy, float Sz);
    void Scale(const Vec3f& S);
    void invScale(float Sx, float Sy, float Sz);
    void invScale(const Vec3f& S);
    void Rotate(float DegAng, Vec3f Axis);
    void invRotate(float DegAng, Vec3f Axis);
    void Frustum(float l, float r, float b, float t, float n, float f);
    void invFrustum(float l, float r, float b, float t, float n, float f);
    void Perspective(float Yfov, float Aspect, float Ndist, float Fdist);
    void invPerspective(float Yfov, float Aspect, float Ndist, float Fdist);
    void Viewport(int WW, int WH);
    void invViewport(int WW, int WH);
    void LookAt(Vec3f Eye, Vec3f LookAtPt, Vec3f ViewUp);
    void invLookAt(Vec3f Eye, Vec3f LookAtPt, Vec3f ViewUp);
    void Viewport2(int WW, int WH);
    void invViewport2(int WW, int WH);
    void Print();
    void CopyInto(float *Mat);
};

#endif
